import numpy as np

from libensemble.message_numbers import (
    MAN_SIGNAL_FINISH,
    TASK_FAILED,
    UNSET_TAG,
    WORKER_DONE,
    WORKER_KILL_ON_ERR,
    WORKER_KILL_ON_TIMEOUT,
)
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_func
from libensemble.specs import input_fields, output_data

__all__ = ["executor_hworld"]

# Alt send values through X
sim_ended_count = 0


def custom_polling_loop(exctr, task, timeout_sec=5.0, delay=0.3):
    import time

    calc_status = UNSET_TAG  # Sim func determines status of libensemble calc - returned to worker

    while task.runtime < timeout_sec:
        time.sleep(delay)

        if exctr.manager_kill_received():
            exctr.kill(task)
            calc_status = MAN_SIGNAL_FINISH  # Worker will pick this up and close down
            print(f"Task {task.id} killed by manager on worker {exctr.workerID}")
            break

        task.poll()
        if task.finished:
            break
        elif task.state == "RUNNING":
            print(f"Task {task.id} still running on worker {exctr.workerID} ....")

        if task.stdout_exists():
            if "Error" in task.read_stdout():
                print(
                    "Found (deliberate) Error in output file - cancelling " f"task {task.id} on worker {exctr.workerID}"
                )
                exctr.kill(task)
                calc_status = WORKER_KILL_ON_ERR
                break

    # After exiting loop
    if task.finished:
        print(f"Task {task.id} done on worker {exctr.workerID}")
        # Fill in calc_status if not already
        if calc_status == UNSET_TAG:
            if task.state == "FINISHED":  # Means finished successfully
                calc_status = WORKER_DONE
            elif task.state == "FAILED":
                calc_status = TASK_FAILED

    else:
        # assert task.state == 'RUNNING', "task.state expected to be RUNNING. Returned: " + str(task.state)
        print(f"Task {task.id} timed out - killing on worker {exctr.workerID}")
        exctr.kill(task)
        if task.finished:
            print(f"Task {task.id} done on worker {exctr.workerID}")
        calc_status = WORKER_KILL_ON_TIMEOUT

    return task, calc_status


@input_fields(["x"])
@output_data([("f", float), ("cstat", int)])
def executor_hworld(H, _, sim_specs, info):
    """
    Tests launching and polling task and exiting on task finish
    """

    exctr = info["executor"]
    cores = sim_specs["user"]["cores"]
    ELAPSED_TIMEOUT = "elapsed_timeout" in sim_specs["user"]

    wait = False
    args_for_sim = "sleep 1"
    calc_status = UNSET_TAG

    batch = len(H["x"])
    H_o = np.zeros(batch, dtype=sim_specs["out"])

    if "six_hump_camel" not in exctr.default_app("sim").full_path:
        global sim_ended_count
        sim_ended_count += 1
        print("sim_ended_count", sim_ended_count, flush=True)

        if ELAPSED_TIMEOUT:
            args_for_sim = "sleep 60"  # Manager kill - if signal received else completes
            timeout = 65.0

        else:
            timeout = 6.0
            launch_shc = False

            if sim_ended_count == 1:
                args_for_sim = "sleep 1"  # Should finish
            elif sim_ended_count == 2:
                args_for_sim = "sleep 1 Error"  # Worker kill on error
            elif sim_ended_count == 3:
                wait = True
                args_for_sim = "sleep 1"  # Should finish
                launch_shc = True
            elif sim_ended_count == 4:
                args_for_sim = "sleep 8"  # Worker kill on timeout
                timeout = 1.0
            elif sim_ended_count == 5:
                args_for_sim = "sleep 2 Fail"  # Manager kill - if signal received else completes

        task = exctr.submit(calc_type="sim", num_procs=cores, app_args=args_for_sim, hyperthreads=True)

        if wait:
            task.wait()
            if not task.finished:
                calc_status = UNSET_TAG
            if task.state == "FINISHED":
                calc_status = WORKER_DONE
            elif task.state == "FAILED":
                calc_status = TASK_FAILED

        else:
            if sim_ended_count >= 2:
                calc_status = exctr.polling_loop(task, timeout=timeout, delay=0.3, poll_manager=True)
                if sim_ended_count == 2 and task.stdout_exists() and "Error" in task.read_stdout():
                    calc_status = WORKER_KILL_ON_ERR
            else:
                task, calc_status = custom_polling_loop(exctr, task, timeout)

    else:
        launch_shc = True
        calc_status = UNSET_TAG

        # Comparing six_hump_camel output, directly called vs. submitted as app
        for i, x in enumerate(H["x"]):
            H_o["f"][i] = six_hump_camel_func(x)
            if launch_shc:
                # Test launching a named app.
                app_args = " ".join(str(val) for val in list(x[:]))
                task = exctr.submit(app_name="six_hump_camel", num_procs=1, app_args=app_args)
                task.wait()
                output = np.float64(task.read_stdout())
                assert np.isclose(H_o["f"][i], output)
                calc_status = WORKER_DONE

    # This is just for testing at calling script level - status of each task
    H_o["cstat"] = calc_status

    return H_o, calc_status
